
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <gnome.h>
#include <glib/gi18n.h>
#include <gconf/gconf-client.h>
#include <gdk/gdkx.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <libhal.h>
#include <libhal-storage.h>

static DBusConnection *dbus_connection;
static LibHalContext *hal_ctx;


static int fds[2];
static int rc;
static gboolean opt_noui = FALSE;
static gboolean opt_block = FALSE;

static void 
notify_parent (gboolean success)
{
	static gboolean already_notified = FALSE;

	if (opt_block) {
		rc = success ? 0 : 1;
	} else {
		if (!already_notified) {
			char buf;
			already_notified = TRUE;
			buf = success ? '1' : '0';
			write (fds[1], &buf, 1);
		}
	}
}

/* #undef SPEW_DEBUG */
#define SPEW_DEBUG

static void
show_error_dialog_no_media (LibHalDrive *drive)
{
	GtkWidget *w;
	if (!opt_noui) {
		w = gtk_message_dialog_new (NULL, 
					    GTK_DIALOG_MODAL,
					    GTK_MESSAGE_ERROR,
					    GTK_BUTTONS_CLOSE,
					    _("Unable to mount media."),
					    libhal_drive_get_udi (drive));
		/* TODO: use icons, text from gnome-vfs, libhal-storage, e.g. s/drive/floppy drive/ */
		gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (w),
							  "There is probably no media in the drive.");
		gtk_dialog_run (GTK_DIALOG (w));
		gtk_widget_unref (w);
	}
}

static void
show_error_dialog_mount (LibHalVolume *volume, const char *error_name, const char *error_detail)
{
	GtkWidget *w;
	if (!opt_noui) {

		/* We need to handle the following errors 
		 *
		 *   org.freedesktop.Hal.Device.Volume.PermissionDenied
		 *   org.freedesktop.Hal.Device.Volume.AlreadyMounted
		 *   org.freedesktop.Hal.Device.Volume.InvalidMountOption
		 *   org.freedesktop.Hal.Device.Volume.FailedToCreateMountpoint
		 *   org.freedesktop.Hal.Device.Volume.UnknownFilesystemType
		 *   org.freedesktop.Hal.Device.Volume.UnknownFailure
		 *
		 * in a sane way.
		 */

		w = gtk_message_dialog_new (NULL, 
					    GTK_DIALOG_MODAL,
					    GTK_MESSAGE_ERROR,
					    GTK_BUTTONS_CLOSE,
					    _("Cannot mount %s"),
					    libhal_volume_get_udi (volume));
		gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (w),
							  "The error was %s : %s",
							  error_name, error_detail);
		gtk_dialog_run (GTK_DIALOG (w));
		gtk_widget_unref (w);
	}
}


/** Integrate a dbus mainloop. 
 *
 *  @param  ctx                 LibHal context
 *  @param  error     		pointer to a D-BUS error object
 *
 *  @return 			TRUE if we connected to the bus
 */
static dbus_bool_t 
hal_mainloop_integration (LibHalContext *ctx, DBusError *error)
{
}

/** Internal HAL initialization function
 *
 * @return			The LibHalContext of the HAL connection or
 *				NULL on error.
 */
static LibHalContext *
do_hal_init (void)
{
	LibHalContext *ctx;
	DBusError error;
	char **devices;
	int nr;
	
	ctx = libhal_ctx_new ();
	if (ctx == NULL) {
		g_warning ("Failed to get libhal context");
		goto error;
	}
	
	dbus_error_init (&error);
	dbus_connection = dbus_bus_get (DBUS_BUS_SYSTEM, &error);
	if (dbus_error_is_set (&error)) {
		g_warning ("Cannot connect to system bus: %s : %s", error.name, error.message);
		dbus_error_free (&error);
		goto error;
	}
	
        dbus_connection_setup_with_g_main (dbus_connection, NULL);	
	libhal_ctx_set_dbus_connection (ctx, dbus_connection);
	
	if (!libhal_ctx_init (ctx, &error)) {
		g_warning ("Failed to initialize libhal context: %s : %s", error.name, error.message);
		dbus_error_free (&error);
		goto error;
	}

	return ctx;

error:
	if (ctx != NULL)
		libhal_ctx_free (ctx);
	return NULL;
}

static gboolean
volume_mount (LibHalVolume *volume, const char *mount_point, const char *fstype, GList *fsoptions)
{
	gboolean ret = FALSE;
	unsigned int n;
	GList *i;
	unsigned int num_options;
	char **options = NULL;
	DBusMessage *msg = NULL;
	DBusMessage *reply = NULL;
	DBusError error;

	if (mount_point == NULL)
		mount_point = "";

	if (fstype == NULL)
		fstype = "";

	msg = dbus_message_new_method_call ("org.freedesktop.Hal", libhal_volume_get_udi (volume),
					    "org.freedesktop.Hal.Device.Volume",
					    "Mount");
	if (msg == NULL) {
		g_warning ("Could not create dbus message for %s", libhal_volume_get_udi (volume));
		goto out;
	}

	num_options = g_list_length (fsoptions);
	options = g_new0 (char *, num_options + 1);

#if defined(SPEW_DEBUG)
	g_message ("mounting %s with mount_point='%s', fstype='%s', num_options=%d", 
		   libhal_volume_get_udi (volume),
		   mount_point, 
		   fstype,
		   num_options);
#endif

	for (i = fsoptions, n = 0; i != NULL; i = g_list_next (i), n++) {
		options[n] = (char *) i->data;
#if defined(SPEW_DEBUG)
		g_message ("  option='%s'", options[n]);
#endif
	}
	options[n] = NULL;
	
	if (!dbus_message_append_args (msg, 
				       DBUS_TYPE_STRING, &mount_point, 
				       DBUS_TYPE_STRING, &fstype,
				       DBUS_TYPE_ARRAY, DBUS_TYPE_STRING, &options, num_options,
				       DBUS_TYPE_INVALID)) {
		g_warning ("Could not append args to dbus message for %s", libhal_volume_get_udi (volume));
		goto out;
	}

	dbus_error_init (&error);
	reply = dbus_connection_send_with_reply_and_block (dbus_connection, msg, -1, &error);
	if (reply = NULL) {
#if defined(SPEW_DEBUG)
		g_warning ("Mount failed for %s: %s : %s\n", libhal_volume_get_udi (volume), error.name, error.message);
#endif
		dbus_error_free (&error);
		goto out;
	}

	if (dbus_error_is_set (&error)) {
#if defined(SPEW_DEBUG)
		g_warning ("Mount failed for %s\n%s : %s\n", libhal_volume_get_udi (volume), error.name, error.message);
#endif

		notify_parent (FALSE);

		show_error_dialog_mount (volume, error.name, error.message);

		dbus_error_free (&error);
		goto out;
	}

	ret = TRUE;

out:
	if (msg != NULL)
		dbus_message_unref (msg);
	if (options != NULL)
		g_free (options);
	if (reply != NULL)
		dbus_message_unref (reply);

	return ret;
}

static gboolean
volume_unmount (LibHalVolume *volume)
{
	gboolean ret = FALSE;
	DBusMessage *msg = NULL;
	DBusMessage *reply = NULL;
	DBusError error;
	char **options = NULL;
	unsigned int num_options = 0;

	msg = dbus_message_new_method_call ("org.freedesktop.Hal", libhal_volume_get_udi (volume),
					    "org.freedesktop.Hal.Device.Volume",
					    "Unmount");
	if (msg == NULL) {
		g_warning ("Could not create dbus message for %s", libhal_volume_get_udi (volume));
		goto out;
	}

	if (!dbus_message_append_args (msg, 
				       DBUS_TYPE_ARRAY, DBUS_TYPE_STRING, &options, num_options,
				       DBUS_TYPE_INVALID)) {
		g_warning ("Could not append args to dbus message for %s", libhal_volume_get_udi (volume));
		goto out;
	}
	
#if defined(SPEW_DEBUG)
	g_message ("unmounting %s", libhal_volume_get_udi (volume));
#endif
	
	dbus_error_init (&error);
	reply = dbus_connection_send_with_reply_and_block (dbus_connection, msg, -1, &error);
	if (reply = NULL) {
		g_warning ("Unmount failed for %s: %s : %s\n", libhal_volume_get_udi (volume), error.name, error.message);
		dbus_error_free (&error);
		goto out;
	}

	if (dbus_error_is_set (&error)) {
		g_warning ("Unmount failed for %s\n%s : %s\n", libhal_volume_get_udi (volume), error.name, error.message);
		dbus_error_free (&error);
		goto out;
	}

	ret = TRUE;

out:
	if (msg != NULL)
		dbus_message_unref (msg);
	if (reply != NULL)
		dbus_message_unref (reply);

	return ret;
}

static gboolean
volume_eject (LibHalVolume *volume)
{
	gboolean ret = FALSE;
	DBusMessage *msg = NULL;
	DBusMessage *reply = NULL;
	DBusError error;
	char **options = NULL;
	unsigned int num_options = 0;

	msg = dbus_message_new_method_call ("org.freedesktop.Hal", libhal_volume_get_udi (volume),
					    "org.freedesktop.Hal.Device.Volume",
					    "Eject");
	if (msg == NULL) {
		g_warning ("Could not create dbus message for %s", libhal_volume_get_udi (volume));
		goto out;
	}

	if (!dbus_message_append_args (msg, 
				       DBUS_TYPE_ARRAY, DBUS_TYPE_STRING, &options, num_options,
				       DBUS_TYPE_INVALID)) {
		g_warning ("Could not append args to dbus message for %s", libhal_volume_get_udi (volume));
		goto out;
	}
	
#if defined(SPEW_DEBUG)
	g_message ("ejecting %s", libhal_volume_get_udi (volume));
#endif
	
	dbus_error_init (&error);
	reply = dbus_connection_send_with_reply_and_block (dbus_connection, msg, -1, &error);
	if (reply = NULL) {
#if defined(SPEW_DEBUG)
		g_warning ("Eject failed for %s: %s : %s\n", libhal_volume_get_udi (volume), error.name, error.message);
#endif
		dbus_error_free (&error);
		goto out;
	}

	if (dbus_error_is_set (&error)) {
#if defined(SPEW_DEBUG)
		g_warning ("Eject failed for %s\n%s : %s\n", libhal_volume_get_udi (volume), error.name, error.message);
#endif
		dbus_error_free (&error);
		goto out;
	}

	ret = TRUE;

out:
	if (msg != NULL)
		dbus_message_unref (msg);
	if (reply != NULL)
		dbus_message_unref (reply);
	return ret;
}


int
main (int argc, char *argv[])
{
	pid_t pid;
	char buf;
	int dev_null_fd;
	LibHalDrive *drive = NULL;
	LibHalVolume *volume = NULL;
	static gboolean opt_verbose = FALSE;
	static gboolean opt_unmount = FALSE;
	static gboolean opt_eject = FALSE;
	static gchar *opt_hal_udi = NULL;
	static gchar *opt_device_file = NULL;
        GError *error = NULL;
        GOptionContext *context = NULL;
        static GOptionEntry entries[] =
                {
                        { "verbose", 'v', 0, G_OPTION_ARG_NONE, &opt_verbose, "Verbose operation", NULL},
                        { "no-ui", 'n', 0, G_OPTION_ARG_NONE, &opt_noui, "Don't show any dialogs", NULL},
                        { "block", 'b', 0, G_OPTION_ARG_NONE, &opt_block, "Allow gnome-mount to block for UI", NULL},
                        { "unmount", 'u', 0, G_OPTION_ARG_NONE, &opt_unmount, "Unmount rather than mount", NULL},
                        { "eject", 'e', 0, G_OPTION_ARG_NONE, &opt_eject, "Eject rather than mount", NULL},
                        { "hal-udi", 'h', 0, G_OPTION_ARG_STRING, &opt_hal_udi, "Mount by HAL UDI", NULL},
                        { "device", 'd', 0, G_OPTION_ARG_STRING, &opt_device_file, "Mount by device file", NULL},
                        { NULL, 0, 0, 0, NULL, NULL, NULL }
                };

	rc = 1;

	bindtextdomain (PACKAGE, GNOMELOCALEDIR);
	bind_textdomain_codeset (PACKAGE, "UTF-8");
	textdomain (PACKAGE);

        context = g_option_context_new ("- GNOME mount");
        g_option_context_add_main_entries (context, entries, PACKAGE);
        g_option_context_add_group (context, gtk_get_option_group (TRUE));
        g_option_context_parse (context, &argc, &argv, &error);

	/* If we want to show dialogs, show that in a child process and let the parent exit immediately
	 * so we don't block
	 */
	if (!opt_block) {
		if (pipe (fds) != 0) {
			fprintf (stderr, "Cannot create pipe\n");
			goto error;
		}
		
		pid = fork();
		switch (pid) {
		case 0:
			/* child; do our real work and notify parent when we know the exit code */
			break;
		case -1:
			/* failure */
			fprintf (stderr, "Cannot fork\n");
			goto error;
			
		default:
			/* parent; wait on rc over pipe from child */
			close (fds[1]);
			read (fds[0], &buf, 1);
			if (buf == '1')
				rc = 0;
			goto error;
		}

		/* close unused descriptor */
		close (fds[0]);

		/* we have to complete daemonize, because otherwise the caller
		 * might wait for the to finish too (e.g.  processes using
		 * libgnomevfs such as Nautilus or any application using the
		 * gnomevfs backend of the filechooser).
		 */
		dev_null_fd = open ("/dev/null", O_RDWR);
		/* ignore if we can't open /dev/null */
		if (dev_null_fd >= 0) {
			/* attach /dev/null to stdout, stdin, stderr */
			dup2 (dev_null_fd, 0);
			dup2 (dev_null_fd, 1);
			dup2 (dev_null_fd, 2);
			close (dev_null_fd);
		}
		setsid ();
	}

	gtk_init (&argc, &argv);

	hal_ctx = do_hal_init ();
	if (hal_ctx == NULL)
		goto out;

	if (opt_hal_udi != NULL) {
		volume = libhal_volume_from_udi (hal_ctx, opt_hal_udi);
	} else if (opt_device_file != NULL) {
		volume = libhal_volume_from_device_file (hal_ctx, opt_device_file);
	} else {
		g_message (_("Use --hal-udi or --device to specify volume"));
		goto out;
	}

	if (volume == NULL) {
		if (opt_hal_udi != NULL) {
			drive = libhal_drive_from_udi (hal_ctx, opt_hal_udi);
		} else if (opt_device_file != NULL) {
			drive = libhal_drive_from_device_file (hal_ctx, opt_device_file);
		}

		if (drive != NULL) {

#if defined(SPEW_DEBUG)
			g_message (_("Drive %s does not contain media."), 
				   opt_device_file != NULL ? opt_device_file : opt_hal_udi);
#endif

			notify_parent (FALSE);

			show_error_dialog_no_media (drive);

		} else {
			/* Silently fail */
#if defined(SPEW_DEBUG)
			g_message (_("Given device '%s' is not a volume or a drive."), 
				   opt_device_file != NULL ? opt_device_file : opt_hal_udi);
#endif
		}

		goto out;
	}


	/* infer from commandline */
	if (strcmp (g_path_get_basename (argv[0]), "gnome-umount") == 0) {
		opt_unmount = TRUE;
	} else if (strcmp (g_path_get_basename (argv[0]), "gnome-eject") == 0) {
		opt_eject = TRUE;
	}

	if (opt_unmount && opt_eject) {
		g_message (_("Cannot unmount and eject simultaneously"));
		goto out;
	}

	if (opt_unmount) {
#if defined(SPEW_DEBUG)
		g_message (_("Unmounting UDI '%s'"), libhal_volume_get_udi (volume));
#endif
		if (volume_unmount (volume))
			rc = 0;
	} else if (opt_eject) {
#if defined(SPEW_DEBUG)
		g_message (_("Ejecting UDI '%s'"), libhal_volume_get_udi (volume));
#endif
		if (volume_eject (volume))
			rc = 0;
	} else {
		const char *fstype;

		fstype = libhal_volume_get_fstype (volume);
		if (fstype != NULL) {
			char uidbuf[64];
			char *mount_point;
			char *fstype_override;
			GList *fsoptions;

			fsoptions = NULL;

			snprintf (uidbuf, sizeof(uidbuf) - 1, "uid=%d", getuid());

			/* TODO: read defaults from gconf */
			if (strcmp (fstype, "vfat") == 0) {
				fsoptions = g_list_append (fsoptions, "shortname=winnt");
				fsoptions = g_list_append (fsoptions, uidbuf);
			} else if (strcmp (fstype, "iso9660") == 0) {
				fsoptions = g_list_append (fsoptions, uidbuf);
			} else if (strcmp (fstype, "udf") == 0) {
				fsoptions = g_list_append (fsoptions, uidbuf);
			}

			/* TODO: read per-volume settings from gconf and adjust fsoptions */

			/* TODO: read property volume.mount.valid_options and adjust fsoptions
			 *       (put into libhal-storage as a convenience function)
			 */

			/* TODO: read mount point from gconf */
			mount_point = NULL;

			/* TODO: read fstype override from gconf */
			fstype_override = NULL;

#if defined(SPEW_DEBUG)
			g_message (_("Mounting UDI '%s'"), libhal_volume_get_udi (volume));
#endif
			if (volume_mount (volume, mount_point, fstype_override, fsoptions))
				rc = 0;

			g_list_free (fsoptions);
		}

	}

	notify_parent (rc == 0);
		
	/*gtk_main ();*/

out:	
 	if (drive != NULL)
		libhal_volume_free (volume);

 	if (volume != NULL)
		libhal_volume_free (volume);

	if (hal_ctx != NULL)
		libhal_ctx_free (hal_ctx);

error:
	if (context != NULL)
		g_option_context_free (context);

	return rc;
}
